@@ -1,18 +1,9 @@
-crew_test("crew_controller_group() method signature compatibility", {
-  x <- crew_controller_callr()
+crew_test("crew_controller_group() method and signature compatibility", {
+  x <- crew_controller_local()
   y <- crew_controller_group(x = x)
-  methods <- c(
-    "validate",
-    "start",
-    "launch",
-    "scale",
-    "collect",
-    "push",
-    "pop",
-    "wait",
-    "summary",
-    "terminate"
-  )
+  common <- intersect(names(x), names(y))
+  methods <- fltr(common, ~is.function(x[[.x]]))
+  methods <- setdiff(methods, "initialize")
   for (method in methods) {
     expect_equal(names(formals(x[[method]])), names(formals(y[[method]])))
   }
@@ -21,27 +12,45 @@ crew_test("crew_controller_group() method signature compatibility", {
 crew_test("crew_controller_group()", {
   skip_on_cran()
   skip_on_os("windows")
-  crew_session_start()
-  a <- crew_controller_callr(
+  a <- crew_controller_local(
     name = "a",
     seconds_idle = 360
   )
-  b <- crew_controller_callr(
+  b <- crew_controller_local(
     name = "b",
-    tasks_max = 1L,
     seconds_idle = 360
   )
   x <- crew_controller_group(a, b)
   expect_null(x$summary())
   on.exit({
     x$terminate()
-    crew_session_terminate()
+    rm(x)
+    gc()
     crew_test_sleep()
   })
   expect_silent(x$validate())
-  expect_false(x$controllers[[1]]$router$listening())
-  expect_false(x$controllers[[2]]$router$listening())
+  for (index in seq_len(2)) {
+    expect_false(x$controllers[[index]]$router$listening())
+  }
   x$start()
+  expect_true(x$empty())
+  expect_true(x$empty(controllers = "a"))
+  expect_true(x$empty(controllers = "b"))
+  for (index in seq_len(2)) {
+    crew_retry(
+      ~x$controllers[[index]]$router$listening(),
+      seconds_interval = 0.001,
+      seconds_timeout = 5
+    )
+  }
+  crew_retry(
+    ~{
+      x$wait(seconds_timeout = 30)
+      TRUE
+    },
+    seconds_interval = 0.001,
+    seconds_timeout = 5
+  )
   s <- x$summary()
   expect_equal(nrow(s), 2L)
   expect_equal(s$controller, c("a", "b"))
@@ -52,7 +61,6 @@ crew_test("crew_controller_group()", {
         "controller",
         "worker_socket",
         "worker_connected",
-        "worker_busy",
         "worker_launches",
         "worker_instances",
         "tasks_assigned",
@@ -77,31 +85,72 @@ crew_test("crew_controller_group()", {
     )
   )
   expect_null(x$pop())
+  # substitute = TRUE # nolint
   x$push(command = ps::ps_pid(), name = "task_pid", controller = "b")
+  expect_false(x$empty())
+  expect_true(x$empty(controllers = "a"))
+  expect_false(x$empty(controllers = "b"))
   x$wait(seconds_timeout = 5)
+  expect_false(x$empty())
+  expect_true(x$empty(controllers = "a"))
+  expect_false(x$empty(controllers = "b"))
   out <- x$pop(scale = FALSE)
+  expect_true(x$empty())
+  expect_true(x$empty(controllers = "a"))
+  expect_true(x$empty(controllers = "b"))
   expect_false(is.null(out))
   expect_equal(out$name, "task_pid")
   expect_equal(out$command, "ps::ps_pid()")
-  expect_equal(
-    out$result[[1]],
-    x$controllers[[2]]$launcher$workers$handle[[1]]$get_pid()
-  )
   expect_true(is.numeric(out$seconds))
   expect_false(anyNA(out$seconds))
   expect_true(out$seconds >= 0)
   expect_true(anyNA(out$error))
-  expect_true(anyNA(out$traceback))
+  expect_true(anyNA(out$trace))
   expect_true(anyNA(out$warnings))
+  pid_out <- out$result[[1]]
+  pid_exp <- x$controllers[[2]]$launcher$workers$handle[[1]]$get_pid()
+  expect_equal(pid_out, pid_exp)
+  # substitute = FALSE # nolint
+  x$push(
+    command = quote(ps::ps_pid()),
+    substitute = FALSE,
+    name = "task_pid2",
+    controller = "a"
+  )
+  x$wait(seconds_timeout = 5)
+  envir <- new.env(parent = emptyenv())
+  crew_retry(
+    ~{
+      envir$out <- x$pop(scale = TRUE)
+      !is.null(envir$out)
+    },
+    seconds_interval = 0.01,
+    seconds_timeout = 10
+  )
+  out <- envir$out
+  expect_false(is.null(out))
+  expect_equal(out$name, "task_pid2")
+  expect_equal(out$command, "ps::ps_pid()")
+  expect_true(is.numeric(out$seconds))
+  expect_false(anyNA(out$seconds))
+  expect_true(out$seconds >= 0)
+  expect_true(anyNA(out$error))
+  expect_true(anyNA(out$trace))
+  expect_true(anyNA(out$warnings))
+  pid_out <- out$result[[1]]
+  pid_exp <- x$controllers[[1]]$launcher$workers$handle[[1]]$get_pid()
+  expect_equal(pid_out, pid_exp)
+  # cleanup
+  handle <- x$controllers[[2]]$launcher$workers$handle[[1]]
   x$terminate()
   for (index in seq_len(2)) {
-    crew_wait(
+    crew_retry(
       ~!x$controllers[[index]]$router$listening(),
       seconds_interval = 0.001,
       seconds_timeout = 5
     )
-    crew_wait(
-      ~(length(x$controllers[[index]]$launcher$active()) == 0L),
+    crew_retry(
+      ~!handle$is_alive(),
       seconds_interval = 0.001,
       seconds_timeout = 5
     )
@@ -111,11 +160,8 @@ crew_test("crew_controller_group()", {
 crew_test("crew_controller_group() select", {
   skip_on_cran()
   skip_on_os("windows")
-  crew_session_start()
-  a <- crew_controller_callr(
-    name = "a"
-  )
-  b <- crew_controller_callr(
+  a <- crew_controller_local(name = "a")
+  b <- crew_controller_local(
     name = "b",
     tasks_max = 1L,
     seconds_idle = 360
@@ -123,7 +169,8 @@ crew_test("crew_controller_group() select", {
   x <- crew_controller_group(a, b)
   on.exit({
     x$terminate()
-    crew_session_terminate()
+    rm(x)
+    gc()
     crew_test_sleep()
   })
   expect_false(a$router$listening())
@@ -140,11 +187,10 @@ crew_test("crew_controller_group() select", {
 crew_test("crew_controller_group() collect", {
   skip_on_cran()
   skip_on_os("windows")
-  crew_session_start()
-  a <- crew_controller_callr(
+  a <- crew_controller_local(
     name = "a"
   )
-  b <- crew_controller_callr(
+  b <- crew_controller_local(
     name = "b",
     tasks_max = 1L,
     seconds_idle = 360
@@ -152,7 +198,8 @@ crew_test("crew_controller_group() collect", {
   x <- crew_controller_group(a, b)
   on.exit({
     x$terminate()
-    crew_session_terminate()
+    rm(x)
+    gc()
     crew_test_sleep()
   })
   expect_silent(x$validate())
@@ -161,12 +208,12 @@ crew_test("crew_controller_group() collect", {
   x$start()
   expect_null(x$pop())
   x$push(command = ps::ps_pid(), name = "task_pid")
-  crew_wait(
+  crew_retry(
     fun = ~{
       x$collect()
       length(x$controllers[[1]]$results) > 0L
     },
-    seconds_interval = 0.001,
+    seconds_interval = 0.01,
     seconds_timeout = 5
   )
   out <- x$pop(scale = FALSE, controllers = "a")
@@ -180,12 +227,11 @@ crew_test("crew_controller_group() collect", {
 crew_test("crew_controller_group() launch method", {
   skip_on_cran()
   skip_on_os("windows")
-  crew_session_start()
-  a <- crew_controller_callr(
+  a <- crew_controller_local(
     name = "a",
     seconds_idle = 360
   )
-  b <- crew_controller_callr(
+  b <- crew_controller_local(
     name = "b",
     tasks_max = 1L,
     seconds_idle = 360
@@ -193,29 +239,31 @@ crew_test("crew_controller_group() launch method", {
   x <- crew_controller_group(a, b)
   on.exit({
     x$terminate()
-    crew_session_terminate()
+    rm(x)
+    gc()
     crew_test_sleep()
   })
   x$start()
   for (index in seq_len(2)) {
-    crew_wait(
-      ~(length(x$controllers[[index]]$launcher$active()) == 0L),
-      seconds_interval = 0.001,
-      seconds_timeout = 5
-    )
+    handle <- x$controllers[[index]]$launcher$workers$handle[[1L]]
+    expect_true(is_crew_null(handle))
   }
   expect_silent(x$launch(n = 1L))
+  handles <- list(
+    x$controllers[[1L]]$launcher$workers$handle[[1L]],
+    x$controllers[[2L]]$launcher$workers$handle[[1L]]
+  )
   for (index in seq_len(2)) {
-    crew_wait(
-      ~(length(x$controllers[[index]]$launcher$active()) == 1L),
+    crew_retry(
+      ~handles[[index]]$is_alive(),
       seconds_interval = 0.001,
       seconds_timeout = 5
     )
   }
   x$terminate()
   for (index in seq_len(2)) {
-    crew_wait(
-      ~(length(x$controllers[[index]]$launcher$active()) == 0L),
+    crew_retry(
+      ~!handles[[index]]$is_alive(),
       seconds_interval = 0.001,
       seconds_timeout = 5
     )
@@ -225,8 +273,7 @@ crew_test("crew_controller_group() launch method", {
 crew_test("crew_controller_group() scale method", {
   skip_on_cran()
   skip_on_os("windows")
-  crew_session_start()
-  a <- crew_controller_callr(
+  a <- crew_controller_local(
     name = "a",
     auto_scale = "one",
     seconds_idle = 360
@@ -234,21 +281,27 @@ crew_test("crew_controller_group() scale method", {
   x <- crew_controller_group(a)
   on.exit({
     x$terminate()
-    crew_session_terminate()
+    rm(x)
+    gc()
     crew_test_sleep()
   })
   x$start()
-  expect_equal(a$launcher$active(), character(0L))
-  a$queue <- list("task")
+  a$push(command = "x", scale = FALSE)
   expect_silent(x$scale())
-  crew_wait(
-    fun = ~identical(length(a$launcher$active()), 1L),
+  crew_retry(
+    ~length(a$launcher$workers$handle) > 0L,
+    seconds_interval = 0.01,
+    seconds_timeout = 5
+  )
+  handle <- a$launcher$workers$handle[[1L]]
+  crew_retry(
+    fun = ~handle$is_alive(),
     seconds_interval = 0.001,
     seconds_timeout = 5
   )
   x$terminate()
-  crew_wait(
-    fun = ~identical(length(a$launcher$active()), 0L),
+  crew_retry(
+    fun = ~!handle$is_alive(),
     seconds_interval = 0.001,
     seconds_timeout = 5
   )
